function [L,U,P] = ilu(M,setup)
% ILU Function
% Sparse Incomplete LU factorization 
% 
% A function computes a sparse lower triangular matrix L and a sparse upper triangular matrix U 
% so the residual matrix R=LU-A satisfies certain constraints, such as having zero entries in 
% some locations. This factorization may be usuful as preconditioners.
%
% 1 - Loadable function: ILU(M) 
% 2 - Loadable function: K = ILU(M)
% 3 - Loadable function: [ L,U ] = ILU(M)
% 4 - Loadable function: ILU(M) 
% 5 - Loadable function: K = ILU(M,setup)
% 6 - Loadable function: [ L,U ] = ILU(M,setup)
% 7 - Loadable function: [ L,U,P ] = ILU(M,setup)
%
% M must be a sparse square matrix.
% Setup is a structure with up to five fields:
% 'type'    -- default = ilutp;
% 'droptol' -- default = 0;
% 'milu'    -- default = off;
% 'udiag'   -- default = 0;
% 'thresh'  -- default = 1;
%
% The possible values of TYPE are:
% 'ilutp'  -- Performs the ILU factorization with threshold and pivoting.
% 'nofill' -- Performs the ILU factorization with zero level of fill-in, know as ILU(0). 
%             Only the milu option is used.
% 'crout'  -- Performs the Crout Version of ILU, know as ILUC. 
%             Only the droptol and milu options are used. 
%
% The value of DROPTOL is a Drop tolerance, is a non-negative scalar.
% Every elements is dropped, replaced by zero, if it is less than the
% droptol times the original norm of the i-th row and 
% droptol times the original norm of the j-th column.
% Setting droptol equal to zero, the complete LU factorization is performed.
% The diagonal entry of U is never dropped. 
% All the entries of L are tested before being scaled by the pivot. 
%
% The possible option of MILU are:
% 'off' -- Performs the standard ILU factorization.
% 'row' -- Performs the modified ILU factorization, know as row-sum. 
%          The diagonal elements are modified to preserve row sums.
%          The product A*e is equal to L*U*e, where e is equal ones(size_of_matrix,1).
% 'col' -- Performs the modified ILU factorization, know as column-sum. 
%          The diagonal elements are modified to preserve column sums.
%          The product e*A is equal to e*L*U, where e is equal ones(1,size_of_matrix).
%
% The possible value of UDIAG are:
% 0 -- Performs the standard ILU factorization.
% 1 -- Any zeros on the diagonal of the upper triangular factor are replaced by 
%      the local drop tolerance, that is equal droptol times norm of i-th row, 
%      this for avoid a singular factor.  
%
% The THRESH is a pivot threshold, is a value between 0 (forces diagonal pivoting) and 1. 
% Performs the pivoting when the maximum magnitude entry in the i-th row times thresh is higher
% than the diagonal entry. 
%
% Numbers 1,2 and 3 compute the nofill factorization.
% Numbers 2 and 5 return "L+U+speye(size_of_matrix)". In a case of 6, the permutation information is lost.
% Numbers 3 and 6 return the lower triangular L and the upper triangular U.
% Number 7 returns the lower triangular L, the upper triangular U and the permutation matrix P.
%
% Fumarola Giovanni Matteo giovanni.fumarola@gmail.com
% Student of Politecnico di Milano
% Project course of "Numerical Analysis".
%
% Reference:
% Saad, Yousef, Iterative Methods for Sparse Linear Systems, PWS Publishing Company, 1996, Chapter 10 ­
% Preconditioning Techniques
% Tzu-Yi Chen, ILUTP_Mem: A Space-Efficient Incomplete LU Preconditioner, A.Lagana et al.(Eds.): 
% ICCSA 2004, LNCS 3046, pp. 20-28, 2004.
  
 
% arguments control
if nargin > 2
   disp('Error: Too many input arguments.')
   return
end

if nargin < 1
   disp('Error: Not enough input arguments.')
   return
end

if nargout > 3
   disp('Error: Too many output arguments.')
   return 
end

if ~isnumeric(M)
   disp('Error: Use ilu(sparse(A),setup).')
   return
end

% sparse control
if issparse(M) == 0 
   disp('Error: Use ilu(sparse(A),setup).')
   return
end

% quadratic control
[ r,c ] = size(M);
if not( r==c )
   disp('Error: Matrix must be square.')
   return
end

if nnz(diag(M)) < r
   disp('Error: Your input matrix has a zero in the diagonal.')
   return
end
if nargin == 2
   try % type value 
      setup2.type = setup.type;
   catch 
      setup2.type = 'ilutp';
   end

   try % droptol value
      setup2.droptol = setup.droptol;
   catch 
      setup2.droptol = 0;
   end

   try % milu value
      setup2.milu = setup.milu;
   catch 
      setup2.milu = 'off';
   end

   try % udiag value
      setup2.udiag = setup.udiag;
   catch 
      setup2.udiag = 0;
   end

   try % thresh value
      setup2.thresh = setup.thresh;
   catch 
      setup2.thresh = 1;
   end

   if setup2.droptol < 0 % droptol is a positive value
      disp('Error: The drop tolerance must be non-negative.');
      return
   end 

   if setup2.udiag ~= 0 & setup2.udiag ~= 1 % udiag is a boolean value
      disp('Error: The option to replace zero diagonals of the upper triangular incomplete LU factor must be either 0 or 1.');
      return
   end

   if setup2.thresh < 0 | setup2.thresh > 1 % thresh is between 0 and 1
      disp('Error: The pivot threshold must be between 0 and 1.');
      return
   end

   if strcmp(setup2.milu,'row') == 0 & strcmp(setup2.milu,'off') == 0 & strcmp(setup2.milu,'col') == 0 
      % only three value aren't forbidden
      disp('Error: The modified option must be row, col or off.');
      return
   end

   if strcmp(setup2.type,'ilutp') == 0 & strcmp(setup2.type,'nofill') == 0 & strcmp(setup2.type,'crout') == 0 
   % only three value aren't forbidden
      disp('Error: The type field must be nofill, crout, ilutp.');
      return
   end

else
   setup2.type = 'nofill';
   setup2.milu = 'off';
end

if nargout == 3 & strcmp(setup2.type,'ilutp') == 0
   disp('Error: The pivoting is forbidden for nofill and crout function. See Help.');
   return
end

% choose condition

if strcmp(setup2.type,'ilutp') == 1
   % ILUTP is performed
   if strcmp(setup2.milu,'off') == 1 
      [ L,U,P ] = ilutp2(M,setup2.thresh,setup2.droptol,setup2.udiag,5,0);                                                                                                                                     elseif strcmp(setup2.milu,'col') == 1 
      %ilutp with milu col
      [ L,U,P ] = ilutp2(M,setup2.thresh,setup2.droptol,setup2.udiag,5,2)   
   else
      %ilutp with milu row
       [ L,U,P ] = ilutp2(M,setup2.thresh,setup2.droptol,setup2.udiag,5,1)   
   end
elseif strcmp(setup2.type,'nofill') == 1
   if strcmp(setup2.milu,'off') == 1
      % NOFILL is performed
      [ L,U ] = nofill(M,0);
   elseif strcmp(setup2.milu,'col') == 1
      %nofill with milu col
     [ L,U ] = nofill(M,2);
   else
      %nofill with milu row
      [ L,U ] = nofill(M,1);
   end 
else
   % ILUC is performed
   if strcmp(setup2.milu,'off') == 1
      % iluc standard
      [ L,U ] = nofill(M,r,0);
   elseif strcmp(setup2.milu,'col') == 1
      % iluc with milu col
      [ L,U ] = nofill(M,r,0); 
   else
      % iluc with milu row
      [ L,U ] = nofill(M,r,0);
   end
end

% only one output 
if nargout <= 1
   L = L+U-speye(r);  
end     

return
end
